@@ -146,7 +146,7 @@ static void print_status(void)
 	fflush (stderr);
 }
 
-static void alarm_intr(int alnum)
+static void alarm_intr(int alnum EXT2FS_ATTR((unused)))
 {
 	signal (SIGALRM, alarm_intr);
 	alarm(1);
@@ -159,7 +159,7 @@ static void alarm_intr(int alnum)
 
 static void *terminate_addr = NULL;
 
-static void terminate_intr(int signo)
+static void terminate_intr(int signo EXT2FS_ATTR((unused)))
 {
 	if (terminate_addr)
 		longjmp(terminate_addr,1);
@@ -188,14 +188,16 @@ static void uncapture_terminate(void)
 	signal (SIGUSR2, SIG_DFL);
 }
 
-static void set_o_direct(int dev, unsigned char *buffer, size_t size)
+static void set_o_direct(int dev, unsigned char *buffer, size_t size,
+			 unsigned long current_block)
 {
 #ifdef O_DIRECT
 	int new_flag = O_DIRECT;
 	int flag;
 	
 	if ((((unsigned long) buffer & (sys_page_size - 1)) != 0) ||
-	    ((size & (sys_page_size - 1)) != 0))
+	    ((size & (sys_page_size - 1)) != 0) ||
+	    ((current_block & ((sys_page_size >> 9)-1)) != 0))
 		new_flag = 0;
 
 	if (new_flag != current_O_DIRECT) {
@@ -214,15 +216,15 @@ static void set_o_direct(int dev, unsigned char *buffer, size_t size)
 static void pattern_fill(unsigned char *buffer, unsigned long pattern,
 			 size_t n)
 {
-	int	i, nb;
+	unsigned int	i, nb;
 	unsigned char	bpattern[sizeof(pattern)], *ptr;
 	
-	if (pattern == ~0) {
+	if (pattern == (unsigned long) ~0) {
 		for (ptr = buffer; ptr < buffer + n; ptr++) {
 			(*ptr) = random() % (1 << (8 * sizeof(char)));
 		}
 		if (s_flag | v_flag)
-			fprintf(stderr, _("Testing with random pattern: "));
+			fputs(_("Testing with random pattern: "), stderr);
 	} else {
 		bpattern[0] = 0;
 		for (i = 0; i < sizeof(bpattern); i++) {
@@ -233,15 +235,17 @@ static void pattern_fill(unsigned char *buffer, unsigned long pattern,
 		}
 		nb = i ? (i-1) : 0;
 		for (ptr = buffer, i = nb; ptr < buffer + n; ptr++) {
-			*ptr = bpattern[i--];
-			if (i < 0)
+			*ptr = bpattern[i];
+			if (i == 0)
 				i = nb;
+			else
+				i--;
 		}
 		if (s_flag | v_flag) {
-			fprintf(stderr, _("Testing with pattern 0x"));
+			fputs(_("Testing with pattern 0x"), stderr);
 			for (i = 0; i <= nb; i++)
 				fprintf(stderr, "%02x", buffer[i]);
-			fprintf(stderr, ": ");
+			fputs(": ", stderr);
 		}
 	}
 }
@@ -255,7 +259,7 @@ static long do_read (int dev, unsigned char * buffer, int try, int block_size,
 {
 	long got;
 
-	set_o_direct(dev, buffer, try * block_size);
+	set_o_direct(dev, buffer, try * block_size, current_block);
 
 	if (v_flag > 1)
 		print_status();
@@ -284,7 +288,7 @@ static long do_write (int dev, unsigned char * buffer, int try, int block_size,
 {
 	long got;
 
-	set_o_direct(dev, buffer, try * block_size);
+	set_o_direct(dev, buffer, try * block_size, current_block);
 
 	if (v_flag > 1)
 		print_status();
@@ -299,8 +303,7 @@ static long do_write (int dev, unsigned char * buffer, int try, int block_size,
 	if (got < 0)
 		got = 0;	
 	if (got & 511)
-		fprintf (stderr,
-			 "Weird value (%ld) in do_write\n", got);
+		fprintf(stderr, "Weird value (%ld) in do_write\n", got);
 	got /= block_size;
 	return got;
 }
@@ -351,7 +354,7 @@ static unsigned int test_ro (int dev, unsigned long last_block,
 		     last_block);
 	}
 	if (t_flag) {
-		fprintf(stderr, _("Checking for bad blocks in read-only mode\n"));
+		fputs(_("Checking for bad blocks in read-only mode\n"), stderr);
 		pattern_fill(blkbuf + blocks_at_once * block_size,
 			     t_patts[0], block_size);
 	}
@@ -360,8 +363,7 @@ static unsigned int test_ro (int dev, unsigned long last_block,
 	currently_testing = from_count;
 	num_blocks = last_block;
 	if (!t_flag && (s_flag || v_flag)) {
-		fprintf(stderr,
-			_("Checking for bad blocks (read-only test): "));
+		fputs(_("Checking for bad blocks (read-only test): "), stderr);
 		if (v_flag <= 1)
 			alarm_intr(SIGALRM);
 	}
@@ -393,6 +395,12 @@ static unsigned int test_ro (int dev, unsigned long last_block,
 		currently_testing += got;
 		if (got == try) {
 			try = blocks_at_once;
+			/* recover page-aligned offset for O_DIRECT */
+			if ( blocks_at_once >= (unsigned long) (sys_page_size >> 9)
+			     && (currently_testing % (sys_page_size >> 9)!= 0))
+				try -= (sys_page_size >> 9)
+					- (currently_testing 
+					   % (sys_page_size >> 9));
 			continue;
 		}
 		else
@@ -435,8 +443,8 @@ static unsigned int test_rw (int dev, unsigned long last_block,
 	flush_bufs();
 
 	if (v_flag) {
-		fprintf(stderr,
-			_("Checking for bad blocks in read-write mode\n"));
+		fputs(_("Checking for bad blocks in read-write mode\n"), 
+		      stderr);
 		fprintf(stderr, _("From block %lu to %lu\n"),
 			 from_count, last_block);
 	}
@@ -467,6 +475,13 @@ static unsigned int test_rw (int dev, unsigned long last_block,
 			currently_testing += got;
 			if (got == try) {
 				try = blocks_at_once;
+				/* recover page-aligned offset for O_DIRECT */
+				if ( blocks_at_once >= (unsigned long) (sys_page_size >> 9)
+				     && (currently_testing % 
+					 (sys_page_size >> 9)!= 0))
+					try -= (sys_page_size >> 9)
+						- (currently_testing 
+						   % (sys_page_size >> 9));
 				continue;
 			} else
 				try = 1;
@@ -481,7 +496,7 @@ static unsigned int test_rw (int dev, unsigned long last_block,
 			fputs(done_string, stderr);
 		flush_bufs();
 		if (s_flag | v_flag)
-			fprintf (stderr, _("Reading and comparing: "));
+			fputs(_("Reading and comparing: "), stderr);
 		num_blocks = last_block;
 		currently_testing = from_count;
 		if (s_flag && v_flag <= 1)
@@ -504,6 +519,14 @@ static unsigned int test_rw (int dev, unsigned long last_block,
 					bb_count += bb_output(currently_testing+i);
 			}
 			currently_testing += got;
+			/* recover page-aligned offset for O_DIRECT */
+			if ( blocks_at_once >= (unsigned long) (sys_page_size >> 9)
+			     && (currently_testing % (sys_page_size >> 9)!= 0))
+				try = blocks_at_once - (sys_page_size >> 9)
+					- (currently_testing 
+					   % (sys_page_size >> 9));
+			else
+				try = blocks_at_once;
 			if (v_flag > 1)
 				print_status();
 		}
@@ -540,9 +563,10 @@ static unsigned int test_nd (int dev, unsigned long last_block,
 	static int num_saved;
 	jmp_buf terminate_env;
 	errcode_t errcode;
-	long buf_used;
-	unsigned int bb_count = 0;
+	unsigned long buf_used;
+	static unsigned int bb_count;
 
+	bb_count = 0;
 	errcode = ext2fs_badblocks_list_iterate_begin(bb_list,&bb_iter);
 	if (errcode) {
 		com_err (program_name, errcode,
@@ -568,19 +592,18 @@ static unsigned int test_nd (int dev, unsigned long last_block,
 
 	flush_bufs();
 	if (v_flag) {
-	    fprintf (stderr,
-		     _("Checking for bad blocks in non-destructive read-write mode\n"));
+	    fputs(_("Checking for bad blocks in non-destructive read-write mode\n"), stderr);
 	    fprintf (stderr, _("From block %lu to %lu\n"), from_count, last_block);
 	}
 	if (s_flag || v_flag > 1) {
-		fprintf(stderr, _("Checking for bad blocks (non-destructive read-write test)\n"));
+		fputs(_("Checking for bad blocks (non-destructive read-write test)\n"), stderr);
 	}
 	if (setjmp(terminate_env)) {
 		/*
 		 * Abnormal termination by a signal is handled here.
 		 */
 		signal (SIGALRM, SIG_IGN);
-		fprintf(stderr, _("\nInterrupt caught, cleaning up\n"));
+		fputs(_("\nInterrupt caught, cleaning up\n"), stderr);
 
 		save_ptr = save_base;
 		for (i=0; i < num_saved; i++) {
@@ -764,11 +787,11 @@ static void check_mount(char *device_name)
 
 	fprintf(stderr, _("%s is mounted; "), device_name);
 	if (force) {
-		fprintf(stderr, _("badblocks forced anyway.  "
-			"Hope /etc/mtab is incorrect.\n"));
+		fputs(_("badblocks forced anyway.  "
+			"Hope /etc/mtab is incorrect.\n"), stderr);
 		return;
 	}
-	fprintf(stderr, _("it's not safe to run badblocks!\n"));
+	fputs(_("it's not safe to run badblocks!\n"), stderr);
 	exit(1);
 }
 
@@ -783,7 +806,7 @@ int main (int argc, char ** argv)
 	char * output_file = NULL;
 	FILE * in = NULL;
 	int block_size = 1024;
-	unsigned long blocks_at_once = 16;
+	unsigned long blocks_at_once = 64;
 	blk_t last_block, from_count;
 	int num_passes = 0;
 	int passes_clean = 0;
@@ -902,7 +925,7 @@ int main (int argc, char ** argv)
 						optarg);
 					exit(1);
 				}
-				if (pattern == ~0)
+				if (pattern == (unsigned long) ~0)
 					pattern = 0xffff;
 				t_patts[t_flag++] = pattern;
 			}
@@ -918,7 +941,7 @@ int main (int argc, char ** argv)
 			  "in read-only mode"));
 			exit(1);
 		}
-		if (t_patts && (t_patts[0] == ~0)) {
+		if (t_patts && (t_patts[0] == (unsigned long) ~0)) {
 			com_err(program_name, 0,
 			_("Random test_pattern is not allowed "
 			  "in read-only mode"));
@@ -962,7 +985,7 @@ int main (int argc, char ** argv)
 	} else from_count = 0;
 	if (from_count >= last_block) {
 	    com_err (program_name, 0, _("bad blocks range: %lu-%lu"),
-		     from_count, last_block);
+		     (unsigned long) from_count, (unsigned long) last_block);
 	    exit (1);
 	}
 	if (w_flag)
